/*
 * $Id$
 *
 ****************************************************************************
 *
 * simulavr - A simulator for the Atmel AVR family of microcontrollers.
 * Copyright (C) 2001, 2002, 2003, 2004  Theodore A. Roth
 * This file:
 * Copyright (C) 2003, 2004 Hermann Kraus
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 ****************************************************************************
 */

#include <config.h>

#include <stdio.h>
#include <stdlib.h>

#include "bitpos.h"

/** \file bitpos.c
  * \brief The file containing all the definitions for the device sepecific
  *  positions of the bits in the \a SFRs.
  *
  * \attention A short HOWTO should be placed here.
  * 
  * \todo I'd like to make structs out of these arrays. But I don't know how to 
  * write a function like bitpos_get_register_mask(), when everything is stored
  * in a stuct.  
  */



/* Note: Maybe it would be better to split this file in device specific files,
   because each device has a lot of specific registers.
   But on the other hand there are a lot of common register. So I don't know 
   what to do. */
   
/* DEFINITIONS FOR AT90S1200 */
static BitPos btab_timsk_at90s1200[LEN_BP_TIMSK_TABLE] = {
    BP_STRUCT( "TIMSK",         LEN_BP_TIMSK_TABLE ),
    BP_STRUCT( "TOIE0",      1 ),
    BP_STRUCT( NULL /* "TOIE1" */,      0xff ),
    BP_STRUCT( NULL /* "TOIE2" */,      0xff ),
    BP_STRUCT( NULL /* "TOIE3" */,      0xff ),
    BP_STRUCT( NULL /* "OCIE0" */,      0xff ),
    BP_STRUCT( NULL /* "OCIE1A" */,     0xff ),
    BP_STRUCT( NULL /* "OCIE1B" */,     0xff ),
    BP_STRUCT( NULL /* "OCIE1C" */,     0xff ),
    BP_STRUCT( NULL /* "OCIE2" */,      0xff ),
    BP_STRUCT( NULL /* "OCIE3A" */,     0xff ),
    BP_STRUCT( NULL /* "OCIE3B" */,     0xff ),
    BP_STRUCT( NULL /* "OCIE3C" */,     0xff ),
    BP_STRUCT( NULL /* "TICIE1" */,     0xff ),
    BP_STRUCT( NULL /* "TICIE3" */,     0xff ),
};

/* DEFINITIONS FOR AT90S4433 */
static BitPos btab_tccr1_at90s4433[] = {
    BP_STRUCT( "TCCR02",         LEN_BP_TCCRX_TABLE ),
    BP_STRUCT( "CS0",        8 ),
    BP_STRUCT( "CS1",        9 ),
    BP_STRUCT( "CS2",        10 ),
    BP_STRUCT( "WGM0/PWM0",  0 ),
    BP_STRUCT( "WGM1/PWM1",  1 ),
    BP_STRUCT( "WGM2/CTC",   11 ),
    BP_STRUCT( NULL /* "WGM3" */,       0xff ),
    BP_STRUCT( "COM0" ,      6 ),
    BP_STRUCT( "COM1",       7 ),
    BP_STRUCT( NULL /* "COMB0" */,      0xff ),
    BP_STRUCT( NULL /* "COMB1" */,      0xff ),
    BP_STRUCT( NULL /* "COMC0" */,      0xff ),
    BP_STRUCT( NULL /* "COMC1" */,      0xff ),
    BP_STRUCT( NULL /* "FOCA" */,       0xff ),
    BP_STRUCT( NULL /* "FOCB" */,       0xff ),
    BP_STRUCT( NULL /* "FOCC" */,       0xff ),
    BP_STRUCT( "ICES",       14 ),
    BP_STRUCT( "ICNC",       15 ),
};


static BitPos btab_timsk_at90s4433[] = {
    BP_STRUCT( "TIMSK",         LEN_BP_TIMSK_TABLE ),
    BP_STRUCT( "TOIE0",      1 ),
    BP_STRUCT( "TOIE1",      7 ),
    BP_STRUCT( NULL /* "TOIE2" */,      0xff ),
    BP_STRUCT( NULL /* "TOIE3" */,      0xff ),
    BP_STRUCT( NULL /* "OCIE0" */,      0xff ),
    BP_STRUCT( "OCIE1",     6 ),
    BP_STRUCT( NULL /* "OCIE1B" */,     0xff ),
    BP_STRUCT( NULL /* "OCIE1C" */,     0xff ),
    BP_STRUCT( NULL /* "OCIE2" */,      0xff ),
    BP_STRUCT( NULL /* "OCIE3A" */,     0xff ),
    BP_STRUCT( NULL /* "OCIE3B" */,     0xff ),
    BP_STRUCT( NULL /* "OCIE3C" */,     0xff ),
    BP_STRUCT( "TICIE1",     3 ),
    BP_STRUCT( NULL /* "TICIE3" */,     0xff ),
};



/* DEFINITIONS FOR AT90S8515 */
static BitPos btab_tccr1_at90s8515[] = {
    BP_STRUCT( "TCCR1",          LEN_BP_TCCRX_TABLE ),
    BP_STRUCT( "CS0",        8 ),
    BP_STRUCT( "CS1",        9 ),
    BP_STRUCT( "CS2",        10),
    BP_STRUCT( "WGM0/PWM0",  0 ),
    BP_STRUCT( "WGM1/PWM1",  1 ),
    BP_STRUCT( "WGM2/CTC",   11),
    BP_STRUCT( "WGM3",       12),
    BP_STRUCT( "COMA0",      6 ),
    BP_STRUCT( "COMA1",      7 ),
    BP_STRUCT( "COMB0",      4 ),
    BP_STRUCT( "COMB1",      5 ),
    BP_STRUCT( NULL /* "COMC0" */,      0xff ),
    BP_STRUCT( NULL /* "COMC1" */,      0xff ),
    BP_STRUCT( NULL /* "FOCA" */,       0xff ),
    BP_STRUCT( NULL /* "FOCB" */,       0xff ),
    BP_STRUCT( NULL /* "FOCC" */,       0xff ),
    BP_STRUCT( "ICES",       14 ),
    BP_STRUCT( "ICNC",       15 ),
};


static BitPos btab_timsk_at90s8515[] = {
    BP_STRUCT( "TIMSK",         LEN_BP_TIMSK_TABLE ),
    BP_STRUCT( "TOIE0",      1 ),
    BP_STRUCT( "TOIE1",      7 ),
    BP_STRUCT( NULL /* "TOIE2" */,      0xff ),
    BP_STRUCT( NULL /* "TOIE3" */,      0xff ),
    BP_STRUCT( NULL /* "OCIE0" */,      0xff ),
    BP_STRUCT( "OCIE1A",     6 ),
    BP_STRUCT( "OCIE1B",     5),
    BP_STRUCT( NULL /* "OCIE1C" */,     0xff ),
    BP_STRUCT( NULL /* "OCIE2" */,      0xff ),
    BP_STRUCT( NULL /* "OCIE3A" */,     0xff ),
    BP_STRUCT( NULL /* "OCIE3B" */,     0xff ),
    BP_STRUCT( NULL /* "OCIE3C" */,     0xff ),
    BP_STRUCT( "TICIE1",     3 ),
    BP_STRUCT( NULL /* "TICIE3" */,     0xff ),
};



/* DEFINITIONS FOR ATMEGA 16 */
static BitPos btab_tccr02_atmega16_128[] = {
    BP_STRUCT( "TCCR02",         LEN_BP_TCCRX_TABLE ),
    BP_STRUCT( "CS0",        0 ),
    BP_STRUCT( "CS1",        1 ),
    BP_STRUCT( "CS2",        2 ),
    BP_STRUCT( "WGM0",       6 ),
    BP_STRUCT( "WGM1",       3 ),
    BP_STRUCT( NULL /* "WGM2" */,       0xff ),
    BP_STRUCT( NULL /* "WGM3" */,       0xff ),
    BP_STRUCT( "COM0",      4 ),
    BP_STRUCT( "COM1",      5 ),
    BP_STRUCT( NULL /* "COMB0" */,      0xff ),
    BP_STRUCT( NULL /* "COMB1" */,      0xff ),
    BP_STRUCT( NULL /* "COMC0" */,      0xff ),
    BP_STRUCT( NULL /* "COMC1" */,      0xff ),
    BP_STRUCT( "FOCA",       7 ),
    BP_STRUCT( NULL /* "FOCB" */,       0xff ),
    BP_STRUCT( NULL /* "FOCC" */,       0xff ),
    BP_STRUCT( NULL /* "ICES" */,       0xff ),
    BP_STRUCT( NULL /* "ICNC" */,       0xff ),
};

static BitPos btab_tccr1_atmega16[] = {
    BP_STRUCT( "TCCR1",          LEN_BP_TCCRX_TABLE ),
    BP_STRUCT( "CS0",        8 ),
    BP_STRUCT( "CS1",        9 ),
    BP_STRUCT( "CS2",        10),
    BP_STRUCT( "WGM0",       0 ),
    BP_STRUCT( "WGM1",       1 ),
    BP_STRUCT( "WGM2",       11),
    BP_STRUCT( "WGM3",       12),
    BP_STRUCT( "COMA0",      6 ),
    BP_STRUCT( "COMA1",      7 ),
    BP_STRUCT( "COMB0",      4 ),
    BP_STRUCT( "COMB1",      5 ),
    BP_STRUCT( NULL /* "COMC0" */,      0xff ),
    BP_STRUCT( NULL /* "COMC1" */,      0xff ),
    BP_STRUCT( "FOCA",       3 ),
    BP_STRUCT( "FOCB",       2 ),
    BP_STRUCT( NULL /* "FOCC" */,       0xff ),
    BP_STRUCT( "ICES",       14 ),
    BP_STRUCT( "ICNC",       15 ),
};

static BitPos btab_timsk_atmega16[] = {
    BP_STRUCT( "TIMSK",         LEN_BP_TIMSK_TABLE ),
    BP_STRUCT( "TOIE0",      0 ),
    BP_STRUCT( "TOIE1",      2 ),
    BP_STRUCT( "TOIE2",      6 ),
    BP_STRUCT( NULL /* "TOIE3" */,      0xff ),
    BP_STRUCT( "OCIE0",      1 ),
    BP_STRUCT( "OCIE1A",     4 ),
    BP_STRUCT( "OCIE1B",     3 ),
    BP_STRUCT( NULL /* "OCIE1C" */,     0xff ),
    BP_STRUCT( "OCIE2",      7 ),
    BP_STRUCT( NULL /* "OCIE3A" */,     0xff ),
    BP_STRUCT( NULL /* "OCIE3B" */,     0xff ),
    BP_STRUCT( NULL /* "OCIE3C" */,     0xff ),
    BP_STRUCT( "TICIE1",     5 ),
    BP_STRUCT( NULL /* "TICIE3" */,     0xff ),
};

static BitPos btab_acsr_atmega16[] = {
    BP_STRUCT( "ACSR",           LEN_BP_ACSR_TABLE ),
    BP_STRUCT( "ACIS0",          0 ),
    BP_STRUCT( "ACIS1",          1 ),
    BP_STRUCT( "ACIC",           2 ),
    BP_STRUCT( "ACIE",           3 ),
    BP_STRUCT( "ACI",            4 ),
    BP_STRUCT( "ACO",            5 ),
    BP_STRUCT( "ACBG",           6 ),
    BP_STRUCT( "ACD",            7 ),
};


/* ATMEGA 128 DEFINITIONS */
static BitPos btab_tccr13_atmega128[] = {
    BP_STRUCT( "TCCR1",          LEN_BP_TCCRX_TABLE ),
    BP_STRUCT( "CS0",        8 ),
    BP_STRUCT( "CS1",        9 ),
    BP_STRUCT( "CS2",        10),
    BP_STRUCT( "WGM0",       0 ),
    BP_STRUCT( "WGM1",       1 ),
    BP_STRUCT( "WGM2",       11),
    BP_STRUCT( "WGM3",       12),
    BP_STRUCT( "COMA0",      6 ),
    BP_STRUCT( "COMA1",      7 ),
    BP_STRUCT( "COMB0",      4 ),
    BP_STRUCT( "COMB1",      5 ),
    BP_STRUCT( "COMC0",      2 ),
    BP_STRUCT( "COMC1",      3 ),
    BP_STRUCT( "FOCA",       23 ),
    BP_STRUCT( "FOCB",       22 ),
    BP_STRUCT( "FOCC",       21 ),
    BP_STRUCT( "ICES",       14 ),
    BP_STRUCT( "ICNC",       15 ),
};

static BitPos btab_timsk_atmega128[] = {
    BP_STRUCT( "TIMSK",         LEN_BP_TIMSK_TABLE ),
    BP_STRUCT( "TOIE0",      0 ),
    BP_STRUCT( "TOIE1",      2 ),
    BP_STRUCT( "TOIE2",      6 ),
    BP_STRUCT( "TOIE3",      10 ),
    BP_STRUCT( "OCIE0",      1 ),
    BP_STRUCT( "OCIE1A",     4 ),
    BP_STRUCT( "OCIE1B",     3 ),
    BP_STRUCT( "OCIE1C",     8 ),
    BP_STRUCT( "OCIE2",      7 ),
    BP_STRUCT( "OCIE3A",     12 ),
    BP_STRUCT( "OCIE3B",     11 ),
    BP_STRUCT( "OCIE3C",     9 ),
    BP_STRUCT( "TICIE1",     5 ),
    BP_STRUCT( "TICIE3",     13 ),
};


/* GENERAL DEFINITIONS */
/* Timers  */
static BitPos btab_tccr__csonly[] = {
    BP_STRUCT( "TCCR02",         LEN_BP_TCCRX_TABLE ),
    BP_STRUCT( "CS0",        0 ),
    BP_STRUCT( "CS1",        1 ),
    BP_STRUCT( "CS2",        2 ),
    BP_STRUCT( NULL /* "WGM0" */,       0xff ),
    BP_STRUCT( NULL /* "WGM1" */,       0xff ),
    BP_STRUCT( NULL /* "WGM2" */,       0xff ),
    BP_STRUCT( NULL /* "WGM3" */,       0xff ),
    BP_STRUCT( NULL /* "COMA0" */,      0xff ),
    BP_STRUCT( NULL /* "COMA1" */,      0xff ),
    BP_STRUCT( NULL /* "COMB0" */,      0xff ),
    BP_STRUCT( NULL /* "COMB1" */,      0xff ),
    BP_STRUCT( NULL /* "COMC0" */,      0xff ),
    BP_STRUCT( NULL /* "COMC1" */,      0xff ),
    BP_STRUCT( NULL /* "FOCA" */,       0xff ),
    BP_STRUCT( NULL /* "FOCB" */,       0xff ),
    BP_STRUCT( NULL /* "FOCC" */,       0xff ),
    BP_STRUCT( NULL /* "ICES" */,       0xff ),
    BP_STRUCT( NULL /* "ICNC" */,       0xff ),
};

/* Watchdog */
static BitPos btab_wdtcr__wdtoe[] = {
    BP_STRUCT( "WDTCR",          LEN_BP_WDTCR_TABLE ),
    BP_STRUCT( "WDP0",       0 ),
    BP_STRUCT( "WDP1",       1 ),
    BP_STRUCT( "WDP2",       2 ),
    BP_STRUCT( "WDE",        3 ),
    BP_STRUCT( "WDTOE",      4 ),
};

static BitPos btab_wdtcr[] = {
    BP_STRUCT( "WDTCR",          LEN_BP_WDTCR_TABLE ),
    BP_STRUCT( "WDP0",       0 ),
    BP_STRUCT( "WDP1",       1 ),
    BP_STRUCT( "WDP2",       2 ),
    BP_STRUCT( "WDE",        3 ),
    BP_STRUCT( NULL /* "WDTOE" */,      0xff ),
};

/* EEProm */
static BitPos btab_eecr__eemwe_eerie[] = {
    BP_STRUCT( "EECR",           LEN_BP_EECR_TABLE ),
    BP_STRUCT( "EERE",       0 ),
    BP_STRUCT( "EEWE",       1 ),
    BP_STRUCT( "EEMWE",      2 ),
    BP_STRUCT( "EERIE",      3 ),
};

static BitPos btab_eecr__eemwe[] = {
    BP_STRUCT( "EECR",           LEN_BP_EECR_TABLE ),
    BP_STRUCT( "EERE",       0 ),
    BP_STRUCT( "EEWE",       1 ),
    BP_STRUCT( "EEMWE",      2 ),
    BP_STRUCT( NULL /* "EERIE" */,      0xff ),
};

static BitPos btab_eecr[] = {
    BP_STRUCT( "EECR",           LEN_BP_EECR_TABLE ),
    BP_STRUCT( "EERE",       0 ),
    BP_STRUCT( "EEWE",       1 ),
    BP_STRUCT( NULL /* "EEMWE" */,      0xff ),
    BP_STRUCT( NULL /* "EERIE" */,      0xff ),
};

BitPosTable bpt_at90s1200 = {
	.tccr0 = btab_tccr__csonly,
	.tccr1 = NULL,
	.tccr2 = NULL,
	.tccr3 = NULL,
	.timsk = btab_timsk_at90s1200,
	.wdtcr = btab_wdtcr,
	.eecr  = btab_eecr,
	.acsr  = btab_acsr_atmega16,
};

BitPosTable bpt_at90s4433 = {
	.tccr0 = btab_tccr__csonly,
	.tccr1 = btab_tccr1_at90s4433,
	.tccr2 = NULL,
	.tccr3 = NULL,
	.timsk = btab_timsk_at90s4433,
	.wdtcr = btab_wdtcr__wdtoe,
	.eecr  = btab_eecr__eemwe_eerie,
	.acsr  = btab_acsr_atmega16,
};

BitPosTable bpt_at90s8515 = {
	.tccr0 = btab_tccr__csonly,
	.tccr1 = btab_tccr1_at90s8515,
	.tccr2 = NULL,
	.tccr3 = NULL,
	.timsk = btab_timsk_at90s8515,
	.wdtcr = btab_wdtcr__wdtoe,
	.eecr  = btab_eecr__eemwe,
	.acsr  = btab_acsr_atmega16,
};

BitPosTable bpt_atmega16 = {
	.tccr0 = btab_tccr02_atmega16_128,
	.tccr1 = btab_tccr1_atmega16,
	.tccr2 = btab_tccr02_atmega16_128,
	.tccr3 = NULL,
	.timsk = btab_timsk_atmega16,
	.wdtcr = btab_wdtcr__wdtoe,
	.eecr  = btab_eecr__eemwe_eerie,
	.acsr  = btab_acsr_atmega16,
};

BitPosTable bpt_atmega128 = {
	.tccr0 = btab_tccr02_atmega16_128,
	.tccr1 = btab_tccr13_atmega128,
	.tccr2 = btab_tccr02_atmega16_128,
	.tccr3 = btab_tccr13_atmega128,
	.timsk = btab_timsk_atmega128,
	.wdtcr = btab_wdtcr__wdtoe,
	.eecr  = btab_eecr__eemwe_eerie,
	.acsr  = btab_acsr_atmega16,
};

/** \brief Create a bitmask for a BitPos-Definition
  * \param btab_register The register for which the bitmask should be created
  * \return The bitmask
  */
uint32_t bitpos_get_register_mask(BitPos *btab_register)
{
  int result=0;
  int i;
  for (i=btab_register[0].bit-1; i>0; i--)
    {
      if (btab_register[i].bit<0xff) {
        result|=_BV(btab_register[i].bit);
      }
    }
  return result;
}




