/*
 * $Id$
 *
 ****************************************************************************
 *
 * simulavr - A simulator for the Atmel AVR family of microcontrollers.
 * Copyright (C) 2001, 2002, 2003, 2004  Theodore A. Roth
 * This file:
 * Copyright (C) 2003, 2004 Hermann Kraus
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 ****************************************************************************
 */

#ifndef SIM_BITPOS_H
#define SIM_BITPOS_H

//#define BIT_DEBUGGING

#ifdef BIT_DEBUGGING
	#define BP_STRUCT(name, bit)  {name, bit}
#else
	#define BP_STRUCT(name, bit)  {bit}
#endif

#define _BV(bit)    1<<(bit)
#define BP_BV(bp_table, bit_name)  _BV(bp_table[bit_name].bit)

//#define BP_BIT_IS_SET(register, bp_table, bit_name) ((register) & bp_table[bit_name].bit)

typedef struct _BitPos BitPos;

struct _BitPos {
#ifdef BIT_DEBUGGING
    char           *name;              /* The name of the flag as described
                                          in the datasheet from ATMEL(only for debuging) */
#endif
    unsigned char  bit;                /* The position in the register */
};

/* Note: If there's a bit for register xxxA and xxxB, but the device only has one
   register xxx (w/o an "A" appended) use the bit for registerX for this register! */
/* Note 2: Registers for the same function are joined in the form
   a) Two Registers:  reg=reg_a+reg_b<<8
   b) Tree Registers: reg=reg_a+reg_b<<8+reg_c<<16 
   This is easier than always checking two or more registers */
/* Note 3: If a register DOESN'T exist there must be NO definition of the bits,
   but in the device's register table you MUST add NULL at its position! */
   
/* I call this tccrx because it can be used for all timers/counters and not
   only for timer/counter0. Only in the individual definitions you have to
   difference between tccr0,tccr1a/b,tccr2, etc. */
enum _bitpos_tccrx {
    TCCRX_LENGTH,
    BP_CS0,                     /* timer/counter clock select bit 0*/
    BP_CS1,                     /* timer/counter clock select bit 1*/
    BP_CS2,                     /* timer/counter clock select bit 2*/
    BP_WGM0,                    /* waveform generation mode   bit 0*/
    BP_WGM1,                    /* waveform generation mode   bit 1*/
    BP_WGM2,                    /* waveform generation mode   bit 2*/
    BP_WGM3,                    /* waveform generation mode   bit 3*/
    BP_COMA0,                   /* compare match output mode  bit 0*/
    BP_COMA1,                   /* compare match output mode  bit 1*/
    BP_COMB0,
    BP_COMB1,
    BP_COMC0,
    BP_COMC1,
    BP_FOCA,                    /* force output compare (port a)*/
    BP_FOCB,                    /* force output compare (port b)*/
    BP_FOCC,                    /* force output compare (port c)*/
    BP_ICES,                    /* input capture edge select */
    BP_ICNC,                    /* input capture noise canceler (4 CKs) */
    LEN_BP_TCCRX_TABLE,
};

enum _bitpos_timsk {
    TIMSK_LENGHT,
    BP_TOIE0,                   /* t/c0 overflow interrupt enable */
    BP_TOIE1,                   /* t/c1 overflow interrupt enable */
    BP_TOIE2,                   /* t/c2 overflow interrupt enable */
    BP_TOIE3,
    BP_OCIE0,                   /* t/c0 output compare match interrupt enable */
    BP_OCIE1A,                  /* t/c1a output compare match interrupt enable */
    BP_OCIE1B,                  /* t/c1b output compare match interrupt enable */
    BP_OCIE1C,
    BP_OCIE2,                   /* t/c2 output compare match interrupt enable */
    BP_OCIE3A,
    BP_OCIE3B,
    BP_OCIE3C,
    BP_TICIE1,                  /* t/c1 input capture interrupt enable */
    BP_TICIE3,
    LEN_BP_TIMSK_TABLE,
};

/* Just the bitnames! Do NOT create an acutal definition as it's the same as timsk */
enum _bitpos_tifr {
    TIFR_LENGHT,
    BP_TOV0,                    /* t/c0 overflow interrupt flag */
    BP_TOV1,                    /* t/c1 overflow interrupt flag */
    BP_TOV2,                    /* t/c2 overflow interrupt flag */
    BP_TOV3,
    BP_OCF0,                    /* t/c0 output compare interrupt flag */
    BP_OCF1A,                   /* t/c1a output compare interrupt flag */
    BP_OCF1B,                   /* t/c1b output compare interrupt flag */
    BP_OCF1C,
    BP_OCF2,                    /* t/c2 output compare interrupt flag */
    BP_OCF3A,
    BP_OCF3B,
    BP_OCF3C,
    BP_ICF1,                    /* t/c1 input capture interrupt flag */
    BP_ICF3,
    LEN_BP_TIFR_TABLE,
};

enum _bitpos_wdtcr {
    WDTCR_LENGTH,
    BP_WDP0,                    /* prescaler bit 0 */
    BP_WDP1,                    /* prescaler bit 1 */
    BP_WDP2,                    /* prescaler bit 2 */
    BP_WDE,                     /* watchdog enable */
    BP_WDTOE,                   /* turn-off enable */
    LEN_BP_WDTCR_TABLE,
};

enum _bitpos_eecr {
    EECR_LENGTH,
    BP_EERE,                    /* eeprom read enable strobe         */
    BP_EEWE,                    /* eeprom write enable strobe        */
    BP_EEMWE,                   /* eeprom master write enable strobe */
    BP_EERIE,                   /* eeprom ready interrupt enable     */
    LEN_BP_EECR_TABLE,
};

enum _bitpos_acsr {
    ACSR_LENGTH,
    BP_ACIS0,               /* AC Interrupt Mode Select 0 */
    BP_ACIS1,               /* AC Interrupt Mode Select 1 */
    BP_ACIC,                /* AC Input Caputre enable */
    BP_ACIE,                /* AC Interrupt Enable */
    BP_ACI,                 /* AC Interrupt flag */
    BP_ACO,                 /* AC Output */
    BP_ACBG,                /* AC Bandgap Select */
    BP_ACD,                 /* AC Disable */
    LEN_BP_ACSR_TABLE,
};

typedef struct _BitPosTable BitPosTable;
struct _BitPosTable {
	BitPos *	tccr0;
	BitPos *	tccr1;
	BitPos *	tccr2;
	BitPos *	tccr3;
	BitPos *	timsk;
	BitPos *	wdtcr;
	BitPos *	eecr;
	BitPos *	acsr;
};

extern BitPosTable bpt_at90s1200;
extern BitPosTable bpt_at90s4433;
extern BitPosTable bpt_at90s8515;
extern BitPosTable bpt_atmega16;
extern BitPosTable bpt_atmega128;

uint32_t bitpos_get_register_mask(BitPos *btab_register);

static inline int bp_bit_is_set(int register_value, BitPos *bp_table, int bitname) 
{
	return register_value & bp_table[bitname].bit;
}
static inline int bp_is_only_bit_set(int register_value, BitPos *bp_table, int bitname) 
{
	return register_value == bp_table[bitname].bit;
}

static inline int bp_dev_has_bit(BitPos *bp_table, int bitname) 
{
	return bp_table[bitname].bit!=0xff;
}

#endif
